/* 
	Définit les différentes classes pour le codage 
	TODO  : UPGRADER vers la libperseus.
*/	
#ifndef _nsPerseusPCC_H_
#define _nsPerseusPCC_H_

#include "nsPerseusUtil.h"

#include "plhash.h"
#include "prlock.h"

#define KMAX_GEN 10
#define MAX_CONT 25
#define MIN_CONT 10

/* 
	la class PCC est la classe basse qui permet le codage/décodage
	et la gestion du bruit.
	TODO : upgrader vers la libperseus
*/
class PCC
{
public:

  PCC();
  ~PCC();
  nsresult generateCode();
  char * toString();
  nsresult code(const nsACString & aData,
          PRUint64 aLength,nsACString & aDataCoded);
  nsresult decode(const nsACString & aDataCoded,
          PRUint64 aLength,nsACString & aData);
  nsresult puncture(char * aStream, PRUint64 aLength, char ** aNewStream,
          PRUint64 *aNewLength);
  nsresult unpuncture(char * aStream, PRUint64 aLength, char ** aNewStream,
          PRUint64 *aNewLength);
  nsresult addNoise(char * aData, PRUint64 aLength);
  void compute(PRUintn * aPoly,PRUint8 aBit,PRUint32 aMemory, PRUint16 *result);
    
  PRUint16 getID();

  //void chiffrement(char * entrée, PRUint64 size_entree, char * sortie, PRUint64 size_sortie);


   PRUint16 mID;
  /*Convolution code parameters*/
  PRUint32 ** mPoly;
  PRUint8 mN;
  PRUint8 mK;
  PRUint8 mM;
  /*Puncutred Matrix parameters*/
  PRUint8 mMatDepth;
  PRUint8 mMatWidth;
  PRUint8 * mMatrix;
  /*Noise parameters*/
  PRUint64 mX0;
  PRUint8 mNumLin;

  PRTime mz;

};

/*
LA classe position permet la gestion des postions dans un noeud dans le treillis 
de viterbi
TODO : A supprimer après l'upgrade vers libperseus
*/
class position
{
public :
    position(PRUint16 input, position * previous);
    ~position();
    nsresult copyMemory(PRUint32 *aMemory, PRUint8 aLength);
    void  deletePosition();
    void setPrevious(position * aPrevious);
    PRUint32 * getMemory();
    PRUint16 getInput();
    position * getPrevious();
private : 
    PRUint16 mInput;
    PRUint32 * mMemory;
    position * mPrevious;
};

/*
LA classe noded permet la gestion des noeuds dans le treillis 
de viterbi
TODO : A supprimer après l'upgrade vers libperseus
*/
class noded
{
public :
    noded(PRInt8 metric, PRUint32 positionNumber, position ** block);
    ~noded();
    void addPosition(PRInt8 aMetric,PRUint16 aInput,PRUint32 *aMemory,
        PRUint8 aMemoryLength, position *aPrevious);
    void replaceBlock(PRInt8 aMetric,PRUint16 aInput,PRUint32 *aMemory,
        PRUint8 aMemoryLength, position *aPrevious);
    void setMetric(PRInt8 aMetric);
    PRUint32 * getMemory(PRUint32 aPosition);
    PRInt8 getMetric();
    PRUint32 getPositionNumber();
    PRBool isEmpty();
    position * getAddressPosition(PRUint32 aPosition);
private :
    position ** mBlock;
    PRInt8 mMetric;
    PRUint32 mPositionNumber;
};

/*
LA classe treillis permet la gestion du treillis de viterbi
TODO : A supprimer après l'upgrade vers libperseus
*/
class treillis
{ 

public:
    treillis(PCC * aPcc, char * aData,   PRUint64 aLength);
    ~treillis();
     void decode(char ** aDataDecoded);
    PRUint32 NumberOfPosition(PRUint64 x, PRUint16 y);
    void addPosition(PRUint64 x, PRUint16 y, PRInt8 aMetric,PRUint16 aInput,
    PRUint32 *aMemory, PRUint8 aMemoryLength, position *aPrevious);
    void setMetric(PRUint64 x, PRUint16 y, PRInt8 aMetric);
    PRInt8 getMetric(PRUint64 x, PRUint16 y);
    PRBool isEmpty(PRUint64 x, PRUint16 y);
    PRUint32 * getMemory(PRUint64 x, PRUint16 y, PRUint32 aPosition);
    position * getAddressPosition(PRUint64 x, PRUint16 y, PRUint32 aPosition);
    void backward(char ** aDataDecoded);

private :
    noded *** mTreillis;
    PRUint64 mWidth;
    PRUint16 mHeight;
    PCC * mPcc;
    char * mData;
    PRUint64 mDataLength;

};

/*
LA classe nsPerseusCode est la classe "haute", elle fait 
l'interface entre les fonctions de firefox et la libraire Perseus.
Implémente une table de hashage pour la sauvegarde des différents codeurs par serveur.
TODO : A upgrader pour le support de la libperseus
*/
class nsPerseusCode  
{
public :
  nsPerseusCode();
  ~nsPerseusCode();
  PRBool hasACode(const nsACString& host);
  PRBool deleteAcode(const nsACString& host);
  nsresult createAcode(const nsACString& host);
  char * toString(const nsACString& host);
  nsresult code(const nsACString& host,const nsACString & aData,
                PRUint64 aLength,
                nsACString & aDataCoded);
  nsresult decode(const nsACString& host,
                  const nsACString & aDataCoded,
                  PRUint64 aLength,
                  nsACString & aData);
  PRUint16 getID(const nsACString &host);

 protected :
  PLHashTable * mTable;

private :
	PRLock *  mLock;

};

#endif //_PerseusPCC_H

