
/********Perseus header**********/
#include "nsPerseusDownloadData.h"
#include "nsPerseusObserver.h"
#include "nsPerseusUploadData.h"

/*****Firefox Core Header********/
#include "nsStringAPI.h"

#include "nsIObserverService.h"

#include "nsIHttpProtocolHandler.h"
#include "nsIHttpChannel.h"
#include "nsITraceableChannel.h"
#include "nsIURI.h"
#include "nsIDOMWindow.h"

#include "prmem.h"

/*****************************************************************************\
 *
 * Class nsPerseusObserver Implementation
 *
 *
 * ***************************************************************************/

nsPerseusObserver::nsPerseusObserver() 
{
  LOG(("[Perseus Extension] - \
       [nsPerseusObserver::nsPerseusObserver] "));
  mRun=PR_FALSE;
}

nsPerseusObserver::~nsPerseusObserver()
{
  LOG(("[Perseus Extension] - \
       [nsPerseusObserver::~nsPerseusObserver]"));

}


NS_IMPL_ISUPPORTS2(nsPerseusObserver,nsIPerseus,nsIObserver)

/*****************************************************************************\
 *
 * nsIPerseus Implementation
 *
 *
 * ***************************************************************************/

/*
	Activation des observateurs :
	un pour observer les envois de trmes.
	un second pour surveiller la reception
*/
NS_IMETHODIMP
nsPerseusObserver::PerseusOn()
{
    LOG(("[Perseus Extension] - [nsPerseusObserver::PerseusOn] Start"));
    nsresult rv ;
    if (!mRun){
        nsCOMPtr<nsIObserverService> obsSvc = 
            do_GetService("@mozilla.org/observer-service;1", &rv);
        NS_ENSURE_SUCCESS(rv, rv);
        rv = obsSvc->AddObserver(this, NS_HTTP_ON_EXAMINE_RESPONSE_TOPIC,
                PR_FALSE);
        NS_ENSURE_SUCCESS(rv, rv);

        rv = obsSvc->AddObserver(this, NS_HTTP_ON_MODIFY_REQUEST_TOPIC,
                PR_FALSE);
        NS_ENSURE_SUCCESS(rv, rv);
        LOG(("[Perseus Extension] - [nsPerseusObserver::PerseusOn] End"));
        mCode = new nsPerseusCode();
        mRun=PR_TRUE;
    }
    return NS_OK;
}
/*
	DsActivation des observateurs :
*/
NS_IMETHODIMP
nsPerseusObserver::PerseusOff()
{
  LOG(("[Perseus Extension] - [nsPerseusObserver::PerseusOff] Start"));

  nsresult rv ;
  nsCOMPtr<nsIObserverService> obsSvc = 
    do_GetService("@mozilla.org/observer-service;1", &rv);
  NS_ENSURE_SUCCESS(rv, rv);
  rv = obsSvc->RemoveObserver(this, NS_HTTP_ON_EXAMINE_RESPONSE_TOPIC);
  NS_ENSURE_SUCCESS(rv, rv);
  rv = obsSvc->RemoveObserver(this, NS_HTTP_ON_MODIFY_REQUEST_TOPIC);
  NS_ENSURE_SUCCESS(rv, rv);

  LOG(("[Perseus Extension] - [nsPerseusObserver::PerseusOff] End"));
  delete(mCode);
  mRun=PR_FALSE;
  return NS_OK;
}

/*****************************************************************************\
 *
 * nsIObserver Implementation
 *
 *
 * ***************************************************************************/

/*
	Fonction appeler lorsque l'une des actions surveiller a lieu.
*/
NS_IMETHODIMP
nsPerseusObserver::Observe(nsISupports *aSubject,
                           const char *aTopic,
                           const PRUnichar *aData)
{
  LOG(("[Perseus Extension] - \
       [nsPerseusObserver::Observer] Start"));
   nsresult rv ;
  nsCOMPtr<nsIChannel> channel = do_QueryInterface(aSubject, &rv);
  NS_ENSURE_SUCCESS(rv, rv);
  
  nsCOMPtr<nsIHttpChannel> httpChannel = 
    do_QueryInterface(aSubject, &rv);
  NS_ENSURE_SUCCESS(rv, rv);

  if (!httpChannel) {
      LOG(("[Perseus Extension] - \
           [nsPerseusObserver::Observer] \
           httpChannel is Empty nothing to do!"));
      return NS_OK;
  }
/*
Cas ou l'observation  t dclench par la reception d'une trame
*/
  if (!strcmp(aTopic, NS_HTTP_ON_EXAMINE_RESPONSE_TOPIC)) {
      LOG(("[Perseus Extension] - \
                  [nsPerseusObserver::Observer] \
                  NS_HTTP_ON_EXAMINE_RESPONSE_TOPIC "));
      nsCOMPtr<nsIURI> uri ;
      rv=channel->GetURI(getter_AddRefs(uri));
      NS_ENSURE_SUCCESS(rv, rv);

      nsCAutoString scheme;
      rv=uri->GetScheme(scheme);
      NS_ENSURE_SUCCESS(rv, rv);
	  /*
		Vrification du protocol
		Only support http
	  */
      if (strcmp(scheme.get(),"http")) {
          LOG(("[Perseus Extension] - \
               [nsPerseusObserver::Observer] \
               Protocol Unknow :%s",scheme.get()));
               PRTime t=PR_Now();
        LOG((" time https  : %lu",t-mhttps));
      }

	  /*
		Vrification de la prsence de l'en-tte Perseus
	  */
      NS_NAMED_LITERAL_CSTRING(head,"Perseus");
      nsCAutoString headValue;
      rv=httpChannel->GetResponseHeader(head,headValue);
      if(NS_FAILED(rv)) {
          LOG(("[Perseus Extension] - \
                      [nsPerseusObserver::Observer] Perseus head not found"));
          return NS_OK;
      }
	   /*
		Vrification de la prsence de l'en-tte Perseus:PCC
	  */
      else if(headValue.Equals("PCC")) {
          LOG(("[Perseus Extension] - \
                      [nsPerseusObserver::Observer] Head Perseus:PCC found!"));

          //Download data and decoding.
          PerseusDownloadData * downloader = new PerseusDownloadData(mz,mCode);

          nsCOMPtr<nsITraceableChannel> trace = 
              do_QueryInterface(aSubject, &rv);
          NS_ENSURE_SUCCESS(rv, rv);

          rv= trace->SetNewListener(downloader,
                  getter_AddRefs(downloader->mListener));
          NS_ENSURE_SUCCESS(rv, rv);
          LOG(("[Perseus Extension] - \
                      [nsPerseusObserver::Observer] Time 1 : %lu, %X",mz,this));

          return NS_OK;
      } else {
          LOG(("[Perseus Extension] - \
                      [nsPerseusObserver::Observer] Perseus Header value unknow!"));
      }
/*
Cas ou l'observation  t dclench par l'envoi d'une trame
*/
  } else if (!strcmp(aTopic, NS_HTTP_ON_MODIFY_REQUEST_TOPIC)) {

      LOG(("[Perseus Extension] - \
           [nsPerseusObserver::Observer] NS_HTTP_ON_MODIFY_REQUEST_TOPIC "));
     
      NS_NAMED_LITERAL_CSTRING(heada,"Perseus");
      nsCAutoString headaValue;
      rv=httpChannel->GetRequestHeader(heada,headaValue);
      if (NS_SUCCEEDED(rv)) {
          LOG(("[Perseus Extension] - \
			   [nsPerseusObserver::Observer] Perseus:%s header found",headaValue.get()));
          return NS_OK;
      }
	  /*
		Si la requte comporte l'en-tte Perseus:check, ne rien faire
		sinon on bloque tout!!
	  */
      /*else if(headaValue.Equals("check")) {
          LOG(("[Perseus Extension] - \
                      [nsPerseusObserver::Observer] Perseus head found"));
          return NS_OK;
      }*/ 
      nsCOMPtr<nsIURI> uri ;
      rv=channel->GetURI(getter_AddRefs(uri));
      NS_ENSURE_SUCCESS(rv, rv);

      if(!uri) {
          LOG((" [Perseus Extension] - \
               [nsPerseusObserver::Observer] URI is Empty "));
          return NS_ERROR_FAILURE;
      }

      nsCAutoString scheme;
      rv=uri->GetScheme(scheme);
      NS_ENSURE_SUCCESS(rv, rv);
	  
	  /*
		Vrification du protocol
		Only support http
	  */
      if (strcmp(scheme.get(),"http")) {
          LOG(("[Perseus Extension] - \
               [nsPerseusObserver::Observer] \
               Protocol Unknow :%s",scheme.get()));
               mhttps=PR_Now();
                LOG((" time https  : %lu",mhttps));
          return NS_ERROR_FAILURE;
      }

      mz=PR_Now();
      LOG(("[Perseus Extension] - \
       [nsPerseusObserver::Observer] Time 1 : %lu %X",mz,this));

	  /*
		On commence le protocole d'change
	  */
      PerseusUploadData * upload = new PerseusUploadData(channel,mCode);
	  /*
		On check si le serveur supporte perseus
	  */
      rv=upload->check(); 

      if (NS_FAILED(rv)) {
		  /*
				Problme ou non support de perseus par le serveur
				on envoi les data en clair.
		  */
          LOG(("[Perseus Extension] - \
               [nsPerseusObserver::Observer] - Upload Set Error !"));
          upload->sendOriginal();
          delete upload;
          return NS_ERROR_FAILURE;
      } else {
          LOG(("[Perseus Extension] - \
               [nsPerseusObserver::Observer] -  Send message encoded !"));
          upload->sendCoded();
          delete upload;
      }
  } else {
      LOG(("[Perseus Extension] - \
           [nsPerseusObserver::Observer] - Observation Unknow "));
  }
   PRTime z;
   z=PR_Now();
   LOG((" TIME 2 : %u, diff : %lu , %X",z,z-mz,this));
   LOG(("[Perseus Extension] - \
       [nsPerseusObserver::Observer] - End Observer"));
  return NS_OK;
}
