#include "nsPerseusUploadData.h"
 
#include "nsStringAPI.h"

#include "nsIUploadChannel.h"
#include "nsIIOService.h"
#include "nsIHttpChannel.h"
#include "nsIURI.h"
#include "nsIRequest.h"
#include "nsIStringStream.h"
#include "nsIMIMEInputStream.h"
#include "nsStringStream.h"

//#include "prmem.h"
#include "plstr.h"
#include "prprf.h"

#define KMAX_HEAD 256
//TODO :
//ADD header and Method in global var

PerseusUploadData::PerseusUploadData( nsIChannel *aChannel, nsPerseusCode *aCode)
:mChannel(aChannel)
{
	LOG(("[Perseus Extension] - "
         "[PerseusUploadData::PerseusUploadData]"));
	mPerseusCode = aCode;
}
 
 
PerseusUploadData::~PerseusUploadData()
{
	LOG(("[Perseus Extension] - "
         "[nsPerseusUploadData::~nsPerseusUploadData]"));
}


/*
	sendParameters : envoi les donnes en HTTPS  host
*/
nsresult
PerseusUploadData::sendParameters(const nsACString& host)
{
    LOG(("[Perseus Extension] - "
         "[PerseusUploadData::sendParameters]"));
    nsresult rv;
	  
    nsCAutoString urlServer("https://");
    urlServer.Append(host);
	LOG(("[Perseus Extension] - "
         "[PerseusUploadData::sendParameters] url created"));

	
    nsCOMPtr<nsIIOService> newio = 
        do_GetService("@mozilla.org/network/io-service;1",&rv);
    NS_ENSURE_SUCCESS(rv, rv);
	LOG(("[Perseus Extension] - "
         "[PerseusUploadData::sendParameters] service obtain"));

    nsCOMPtr<nsIURI> uri ;
    rv=newio->NewURI(urlServer,nsnull,nsnull,getter_AddRefs(uri));
    NS_ENSURE_SUCCESS(rv, rv);
	LOG(("[Perseus Extension] - "
         "[PerseusUploadData::sendParameters] io created"));

    nsCOMPtr<nsIChannel> channel ;
	
	ToNewCString(host);	
	rv=newio->NewChannelFromURI(uri,getter_AddRefs(channel));

	NS_ENSURE_SUCCESS(rv, rv);
	LOG(("[Perseus Extension] - "
         "[PerseusUploadData::sendParameters] chan created"));

    nsCOMPtr<nsIStringInputStream> stream = 
        do_GetService(NS_STRINGINPUTSTREAM_CONTRACTID,&rv) ;
    NS_ENSURE_SUCCESS(rv, rv);
	LOG(("[Perseus Extension] - "
         "[PerseusUploadData::sendParameters] Stream service obtain"));

	char * parameters = mPerseusCode->toString(host);
	LOG(("[Perseus Extension] - "
		 "[PerseusUploadData::sendParameters] Test parameters"));
	if (!parameters) {
		LOG(("[Perseus Extension] - "
             "[PerseusUploadData::sendParameters] Error to set Parameters in string format"));
        return NS_ERROR_OUT_OF_MEMORY;
	}
    
	stream->SetData((const char *)parameters,PL_strlen(parameters));
    NS_Free(parameters);
	
    LOG(("[Perseus Extension] - "
         "[PerseusUploadData::sendParameters] Prepare Https Channel"));
    nsCOMPtr<nsIUploadChannel> uploadchannel = 
        do_QueryInterface(channel, &rv);
    NS_ENSURE_SUCCESS(rv, rv);

    nsCOMPtr<nsISeekableStream> seekableStream =
        do_QueryInterface(stream, &rv);
    NS_ENSURE_SUCCESS(rv, rv);

    rv=seekableStream->Seek(0,0);
    NS_ENSURE_SUCCESS(rv, rv);

    nsCOMPtr<nsIInputStream> streamInput=  
        do_QueryInterface(stream, &rv);
    NS_ENSURE_SUCCESS(rv, rv);

    LOG(("[Perseus Extension] - "
         "[PerseusUploadData::sendParameters] Set Header"));
    NS_NAMED_LITERAL_CSTRING(content, "perseus-init");
    rv= uploadchannel->SetUploadStream(streamInput,content,-1);
    NS_ENSURE_SUCCESS(rv, rv);

    nsCOMPtr<nsIHttpChannel> httpChannel =
        do_QueryInterface(channel, &rv);
    NS_ENSURE_SUCCESS(rv, rv);
    
    NS_NAMED_LITERAL_CSTRING(PerseusHead,"Perseus");
    NS_NAMED_LITERAL_CSTRING(PerseusValue,"init");

    rv=httpChannel->SetRequestHeader(PerseusHead,PerseusValue,PR_FALSE);
    NS_ENSURE_SUCCESS(rv, rv);
    LOG(("[Perseus Extension] - "
         "[PerseusUploadData::sendParameters] Get ID"));
    
    NS_NAMED_LITERAL_CSTRING(IDHead,"Perseus_id");
    PRUint16 id=mPerseusCode->getID(host);
    char* idstring = PR_sprintf_append(0,"%u",id);
    nsCAutoString IDValue;
    IDValue.Assign(idstring);
	
    rv=httpChannel->SetRequestHeader(IDHead,IDValue,PR_FALSE);
    NS_ENSURE_SUCCESS(rv, rv);
    NS_Free(idstring);
    
    NS_NAMED_LITERAL_CSTRING(meth,"POST");
    rv=httpChannel->SetRequestMethod(meth);
    NS_ENSURE_SUCCESS(rv, rv);
    LOG(("[Perseus Extension] - "
         "[PerseusUploadData::sendParameters] Parameters begin sent"));

    nsCOMPtr<nsIInputStream> inStream;
    rv=channel->Open(getter_AddRefs(inStream));
    NS_ENSURE_SUCCESS(rv, rv);
    LOG(("[Perseus Extension] - "
         "[PerseusUploadData::sendParameters] Parameters sent"));

    PRUint32 status;

    rv=httpChannel->GetResponseStatus(&status);
    NS_ENSURE_SUCCESS(rv, rv);
    if(status != 200)
        return NS_ERROR_FAILURE;

    NS_NAMED_LITERAL_CSTRING(head,"Perseus");
    nsCAutoString headValue;
    rv=httpChannel->GetResponseHeader(head,headValue);
    NS_ENSURE_SUCCESS(rv, rv);
    if (!headValue.Equals("init-ack")) 
        return NS_ERROR_FAILURE;
    LOG(("[Perseus Extension] - "
         "[PerseusUploadData::sendParameters] Parameters sent ok"));

    return NS_OK;
}

/*
	Check envoi soit :
		* une requte check si le serveur n'est pas connu
			i.e. aucune donnes n'a t chang auparavant
		* une requte known si le serveur est connu
			i.e. un codeur est lie au serveur
*/
nsresult 
PerseusUploadData::check()
{
    LOG(("[Perseus Extension] - \
                [PerseusUploadData::check]"));
    nsresult rv;

    //First we check if the http frame contains some data
    nsCOMPtr<nsIHttpChannel> httpChannel = 
        do_QueryInterface(mChannel, &rv);
    NS_ENSURE_SUCCESS(rv, rv);

    nsCOMPtr<nsIURI> Uri ;
    rv=mChannel->GetURI(getter_AddRefs(Uri));
    NS_ENSURE_SUCCESS(rv, rv);
    
    rv=httpChannel->GetRequestMethod(mMethod);
    NS_ENSURE_SUCCESS(rv, rv);
    if (mMethod.Equals("POST")) {
        rv=readDataPost();
        LOG(("[Perseus Extension] - "
             "[PerseusUploadData::check] Method POST, Read Data"));
        NS_ENSURE_SUCCESS(rv, rv);
    } else if (mMethod.Equals("GET")) {
        rv=readDataGet(Uri);
        LOG(("[Perseus Extension] - "
             "[PerseusUploadData::check] Method GET, Read Data"));
        NS_ENSURE_SUCCESS(rv, rv);
    }

    LOG(("[Perseus Extension] - "
          "[PerseusUploadData::check] read check OK, data :\n %s",mData.get()));

    //nsCAutoString host;
    rv=Uri->GetHost(mHost);
    NS_ENSURE_SUCCESS(rv, rv);

    NS_NAMED_LITERAL_CSTRING(IDHead,"Perseus_id");
    NS_NAMED_LITERAL_CSTRING(PerseusHead,"Perseus");
    nsCAutoString PerseusValue;

    /*
     *Test if we known the server.
     */
    if (!mPerseusCode->hasACode(mHost)) {
        //The server is unknown

        //Check if the server has Perseus

        //URL creation
        nsCAutoString urlServer("http://");
        urlServer.Append(ToNewCString(mHost));

        nsCOMPtr<nsIIOService> newio = 
            do_GetService("@mozilla.org/network/io-service;1",&rv);
        NS_ENSURE_SUCCESS(rv, rv);

        nsCOMPtr<nsIURI> uri ;
        rv=newio->NewURI(urlServer,nsnull,nsnull,getter_AddRefs(uri));
        NS_ENSURE_SUCCESS(rv, rv);

        //Http channel creation
        nsCOMPtr<nsIChannel> channel ;
        rv=newio->NewChannelFromURI(uri,getter_AddRefs(channel));
        NS_ENSURE_SUCCESS(rv, rv);


        nsCOMPtr<nsIStringInputStream> stream = 
            do_GetService(NS_STRINGINPUTSTREAM_CONTRACTID,&rv) ;
        NS_ENSURE_SUCCESS(rv, rv);

        nsCOMPtr<nsIHttpChannel> ahttpChannel = 
            do_QueryInterface(channel, &rv);
        NS_ENSURE_SUCCESS(rv, rv);
        NS_NAMED_LITERAL_CSTRING(headValue,"check");
        rv=ahttpChannel->SetRequestHeader(PerseusHead,headValue,PR_FALSE);
        NS_ENSURE_SUCCESS(rv, rv);
        LOG(("[Perseus Extension] - "
             "[PerseusUploadData::check] Send head check to : %s",urlServer.get()));

        NS_NAMED_LITERAL_CSTRING(meth,"Head");
        rv=ahttpChannel->SetRequestMethod(meth);
        NS_ENSURE_SUCCESS(rv, rv);

        nsCOMPtr<nsIRequest> request = 
            do_QueryInterface(channel, &rv);
        nsLoadFlags flag = nsIRequest::LOAD_BYPASS_CACHE;
       request->SetLoadFlags(flag);

        nsCOMPtr<nsIInputStream> inStream;
        //If the server doesn't reply => BUG : we are blocked!
        rv=channel->Open(getter_AddRefs(inStream));
        NS_ENSURE_SUCCESS(rv, rv);

        PRUint32 status;

        rv=ahttpChannel->GetResponseStatus(&status);
        NS_ENSURE_SUCCESS(rv, rv);
        if(status != 200)
            return NS_ERROR_FAILURE;
        LOG(("[Perseus Extension] - "
             "[PerseusUploadData::check] Send head check OK"));

        rv=ahttpChannel->GetResponseHeader(PerseusHead,PerseusValue);
        NS_ENSURE_SUCCESS(rv, rv);

        if (!PerseusValue.Equals("check-ack")) 
        {
            LOG(("[Perseus Extension] - "
                 "[PerseusUploadData::check] Send head check-ack not OK, %s",PerseusValue.get()));

            return NS_ERROR_FAILURE;
        }
        
		rv=mPerseusCode->createAcode(mHost);
        NS_ENSURE_SUCCESS(rv, rv);
       // mPerseusCode->toString(host);
		rv=sendParameters(mHost);
        NS_ENSURE_SUCCESS(rv, rv);
    }

    /*********************************/
    /*******************************/
    else  {
        LOG(("[Perseus Extension] - \
                    [PerseusUploadData::check] Code is known"));

        //Ask to the server if he knows the code
        //Send a known request
        //URL creation
        nsCAutoString urlServer("http://");
        urlServer.Append(ToNewCString(mHost));

        nsCOMPtr<nsIIOService> newio = 
            do_GetService("@mozilla.org/network/io-service;1",&rv);
        NS_ENSURE_SUCCESS(rv, rv);

        nsCOMPtr<nsIURI> uri ;
        rv=newio->NewURI(urlServer,nsnull,nsnull,getter_AddRefs(uri));
        NS_ENSURE_SUCCESS(rv, rv);

        //Http channel creation
        nsCOMPtr<nsIChannel> channel ;
        rv=newio->NewChannelFromURI(uri,getter_AddRefs(channel));
        NS_ENSURE_SUCCESS(rv, rv);


        nsCOMPtr<nsIStringInputStream> stream = 
            do_GetService(NS_STRINGINPUTSTREAM_CONTRACTID,&rv) ;
        NS_ENSURE_SUCCESS(rv, rv);

        nsCOMPtr<nsIHttpChannel> knownhttpChannel = 
            do_QueryInterface(channel, &rv);
        NS_ENSURE_SUCCESS(rv, rv);
        NS_NAMED_LITERAL_CSTRING(head1Value,"known");
        rv=knownhttpChannel->SetRequestHeader(PerseusHead,head1Value,PR_FALSE);
        NS_ENSURE_SUCCESS(rv, rv);

        PRUint16 id=mPerseusCode->getID(mHost);
        char* idstring = PR_sprintf_append(0,"%u",id);
        nsCAutoString IDValue;
        IDValue.Assign(idstring);
        NS_Free(idstring);
        rv=knownhttpChannel->SetRequestHeader(IDHead,IDValue,PR_FALSE);
        NS_ENSURE_SUCCESS(rv, rv);


        LOG(("[Perseus Extension] - \
                    [PerseusUploadData::check] Send head known to : %s"
                    ,urlServer.get()));

        NS_NAMED_LITERAL_CSTRING(meth,"HEAD");
        rv=knownhttpChannel->SetRequestMethod(meth);
        NS_ENSURE_SUCCESS(rv, rv);
        
        nsCOMPtr<nsIRequest> request = 
            do_QueryInterface(channel, &rv);
      // nsLoadFlags flag = nsIRequest::LOAD_BYPASS_CACHE;
       //request->SetLoadFlags(flag);

        nsCOMPtr<nsIInputStream> inStream;
        //If the server doesn't reply => BUG : we are blocked!
        rv=channel->Open(getter_AddRefs(inStream));
        NS_ENSURE_SUCCESS(rv, rv);

        PRUint32 status;

        rv=knownhttpChannel->GetResponseStatus(&status);
        NS_ENSURE_SUCCESS(rv, rv);
        if (status != 200)
            return NS_ERROR_FAILURE;
        LOG(("[Perseus Extension] - "
             "[PerseusUploadData::check] Send head known OK"));

        NS_NAMED_LITERAL_CSTRING(PersHead,"perseus");
        nsCAutoString PersValue;
        rv=knownhttpChannel->GetResponseHeader(PersHead,PersValue);
        NS_ENSURE_SUCCESS(rv, rv);
        if (PersValue.Equals("known-yes"))  {
        LOG(("[Perseus Extension] - "
             "[PerseusUploadData::check] known-yes"));
            //Nothing to do
        }
		/*
			Le serveur ne connat plus les paramtres, cration et envoi de nouveaux
		*/
        else if (PersValue.Equals("known-no")) {
			LOG(("[Perseus Extension] - \
                    [PerseusUploadData::check] known-no"));
            //Erase parameters
            if (!mPerseusCode->deleteAcode(mHost))
                return NS_ERROR_FAILURE;
            //Create new parameters
            rv=mPerseusCode->createAcode(mHost);
            NS_ENSURE_SUCCESS(rv, rv);
            //Send the new parameters
            rv=this->sendParameters(mHost);
            NS_ENSURE_SUCCESS(rv, rv);
        }
        else return NS_ERROR_FAILURE;
    }

    return NS_OK;
}

nsresult
PerseusUploadData::sendOriginal()
{
	return NS_OK;
}

/*
 Envoi les donnes sous forme codes
	*La fonction code les donnes et place les en-ttes
*/
nsresult
PerseusUploadData::sendCoded()
{
    LOG(("[Perseus Extension] - \
                [PerseusUploadData::sendCoded]"));
    nsresult rv;
    nsCAutoString encoded;

    nsCOMPtr<nsIURI> Uri ;
    rv=mChannel->GetURI(getter_AddRefs(Uri));
    NS_ENSURE_SUCCESS(rv, rv);

    nsCAutoString host;
    rv=Uri->GetHost(host);
    NS_ENSURE_SUCCESS(rv, rv);
    LOG(("[Perseus Extension] - \
		 [PerseusUploadData::sendCoded] Coding Start : %s",mData.get()));

    rv=mPerseusCode->code(host,mData,mData.Length(),encoded);
    LOG(("[Perseus Extension] - \
		 [PerseusUploadData::sendCoded] Coding Finish : %s",encoded.get()));

	nsCAutoString toto;
	mPerseusCode->decode(host,encoded,encoded.Length(),toto);
	LOG(("decoded : %s",toto.get()));
	//Method Get
    if (mMethod.Equals("GET")) {
        nsCAutoString data;
        rv=Uri->GetPath(data);
        NS_ENSURE_SUCCESS(rv, rv);
        if(data.IsEmpty()) {
            return NS_ERROR_FAILURE;
        }
        if (data.FindChar('?') != -1) {
            data.Cut(data.FindChar('?')+1,data.Length());
            data.Append(encoded.get());
            rv=Uri->SetPath(data);
            NS_ENSURE_SUCCESS(rv, rv);
        } else 
            return NS_ERROR_FAILURE;
        //return NS_OK;
    }
    else if (mMethod.Equals("POST")) {
        nsCOMPtr<nsIStringInputStream> stream = 
            do_GetService(NS_STRINGINPUTSTREAM_CONTRACTID,&rv);
        NS_ENSURE_SUCCESS(rv, rv);

        rv=stream->SetData((const char *)encoded.get(),encoded.Length());
        NS_ENSURE_SUCCESS(rv, rv);

        nsCOMPtr<nsIUploadChannel> uploadchannel = 
            do_QueryInterface(mChannel, &rv);
        NS_ENSURE_SUCCESS(rv, rv);


        nsCOMPtr<nsISeekableStream> seekableStream =
            do_QueryInterface(stream, &rv);
        NS_ENSURE_SUCCESS(rv, rv);

        rv=seekableStream->Seek(0,0);
        NS_ENSURE_SUCCESS(rv, rv);

        nsCOMPtr<nsIInputStream> streamInput=  do_QueryInterface(stream, &rv);
        NS_ENSURE_SUCCESS(rv, rv);

        rv=uploadchannel->SetUploadStream(streamInput,mContent,-1);
        NS_ENSURE_SUCCESS(rv, rv);
    }

    nsCOMPtr<nsIHttpChannel> httpChannel = do_QueryInterface(mChannel, &rv);
    NS_ENSURE_SUCCESS(rv, rv);

    NS_NAMED_LITERAL_CSTRING(head,"Perseus");
    NS_NAMED_LITERAL_CSTRING(headVal,"pcc");
    rv=httpChannel->SetRequestHeader(head,headVal,PR_FALSE);
    NS_ENSURE_SUCCESS(rv, rv);

    NS_NAMED_LITERAL_CSTRING(IDHead,"Perseus_id");    
    PRUint16 id=mPerseusCode->getID(host);
    char* idstring = PR_sprintf_append(0,"%u",id);
    nsCAutoString IDValue;
    IDValue.Assign(idstring);
    NS_Free(idstring);

    rv=httpChannel->SetRequestHeader(IDHead,IDValue,PR_FALSE);
    NS_ENSURE_SUCCESS(rv, rv);

    rv=httpChannel->SetRequestMethod(mMethod);
    NS_ENSURE_SUCCESS(rv, rv);
    return NS_OK;
}


/* Return NS_OK if the uri contains data
 * return NS_ERROR_FAILURE else
 */
/*
	Lis et stocke les donnes contenues dans la mthode GET
*/
nsresult
PerseusUploadData::readDataGet(nsIURI *uri)
{
    LOG(("[Perseus Extension] - \
                [nsPerseusUploadData::readDataGet] Get Method Detected"));

    nsresult rv ;
    nsCAutoString data;
    rv=uri->GetPath(data);
    NS_ENSURE_SUCCESS(rv, rv);
    if(data.IsEmpty()) {
        return NS_ERROR_FAILURE;
    }
    if (data.FindChar('?') != -1) {
        data.Cut(0,data.FindChar('?')+1);
    } else 
        return NS_ERROR_FAILURE;
    mData.Assign(data);
    return NS_OK;
}

/*
	Lis et stocke les donnes contenues dans la mthode POST
*/
nsresult
PerseusUploadData::readDataPost()
{
  
	nsresult rv;
	nsCOMPtr<nsIStringInputStream> streamS;
	nsCOMPtr<nsIMIMEInputStream> streamM;
	nsCOMPtr<nsISeekableStream> seekableStream;
	nsCOMPtr<nsIInputStream> stream;
	
	LOG(("[Perseus Extension] - \
       [nsPerseusUploadData::readDataPost] Post Method Detected"));


  nsCOMPtr<nsIUploadChannel> uploadChannel = do_QueryInterface(mChannel, &rv);
  NS_ENSURE_SUCCESS(rv, rv);
  
  if (!uploadChannel) {
      LOG(("[Perseus Extension] - \
           [nsPerseusUploadData::readData] uploadChannel is empty : Error"));
      rv = NS_ERROR_FAILURE;
	  goto end;
  }

  rv=uploadChannel->GetUploadStream(getter_AddRefs(stream));
  if (NS_FAILED(rv)) {
	  goto end;
  }

  mStream=stream;  
  if (!mStream) {
      LOG(("[Perseus Extension] - \
           [nsPerseusUploadData::readData] Input Stream empty nothing to do"));
      rv = NS_ERROR_FAILURE;
	  goto end;
  }

  seekableStream =  do_QueryInterface(mStream, &rv);
  if (NS_FAILED(rv)) {
	  goto end;
  }

  rv=seekableStream->Seek(0,0);
  if (NS_FAILED(rv)) {
	  goto end;
  }

  rv=stream->Available(&mSizeBody);
  if (NS_FAILED(rv)) {
	  goto end;
  }

  LOG(("[Perseus Extension] - \
       [nsPerseusUploadData::readData] InputStream size : %lu",mSizeBody)); 

    /*
     * Test if the stream is a Mime Stream or a String Stream
     * if Mime Stream : Content-type and content-length are in the upload channel
     * else they are in the channel header 
     */
  streamM =  do_QueryInterface(mStream, &rv);
  if (NS_SUCCEEDED(rv)) {
	  rv = readContentMime();
	  if (NS_FAILED(rv)) {
		  goto end;
	  }
  }
  else {
		streamS = do_QueryInterface(mStream, &rv);
		if (NS_SUCCEEDED(rv)) {
			rv = readContentString();
			if (NS_FAILED(rv)) {
			 goto end;
			};
		} else {
			LOG(("[Perseus Extension] - "
                 "[nsPerseusUploadData::readData] InputStream is unknow"));
			rv = NS_ERROR_FAILURE;
			goto end;
		}
  }
  
  LOG((" [Perseus Extension] - \
              [nsPerseusUploadData::readData] Content type : %s",mContent.get()));
  
  rv = readBody(seekableStream);
  if (NS_FAILED(rv)) {
	  goto end;
  }

end:
  return rv;
}


nsresult
PerseusUploadData::readBody(nsISeekableStream* seekableStream) {
  PRInt64 body;
  PRUint32 read;
  nsresult rv ;
  
  seekableStream->Tell(&body);

  char *data= (char *) NS_Alloc( sizeof(char) * (mSizeBody+2) );
  if (!data) {
      rv = NS_ERROR_OUT_OF_MEMORY;
	  goto end;
  }
  
  rv=mStream->Read(data,mSizeBody,&read);
  if (NS_FAILED(rv)) {
	  goto end1;
  }
  data[read]='\0'; 
  LOG(("[Perseus Extension] - "
       "[nsPerseusUploadData::readData] mData read "
       "\n%s, read : %lu",data,read));
  rv=seekableStream->Seek(0,0);
  if (NS_FAILED(rv)) {
	  goto end1;
  }

  rv = NS_OK;
  mData.Assign(data);
end1:
  NS_Free(data);
end:
  return rv;
}

nsresult
PerseusUploadData::readContentMime()
{
	  nsresult rv ;
	  char ** dataHead;
	  PRUint32 read;
      PRUint32 j=0, i=0;
	char c;
	PRUint16 tmp=0,tmp2=0;
char * content;

      LOG(("[Perseus Extension] - \
           [nsPerseusUploadData::readContent] InputStream is a MimeInputStream"));
      //We must read Content-type and content-lenght before the data
      dataHead =(char **) NS_Alloc(sizeof(char*) * 2);
      if (!dataHead) {
          rv = NS_ERROR_OUT_OF_MEMORY;
		  goto end;
      }
		
      dataHead[0]= (char *) NS_Alloc( sizeof(char) * KMAX_HEAD );
      if (!dataHead[0]) {
          rv = NS_ERROR_OUT_OF_MEMORY;
		  goto end1;
      }

      dataHead[1]= (char *) NS_Alloc( sizeof(char) * KMAX_HEAD );
      if (!dataHead[1]) {
          rv = NS_ERROR_OUT_OF_MEMORY;
		  goto end2;
      }

      for (j=0;j<2;j++) {
          i=0;
          while(i<KMAX_HEAD) {
              rv=mStream->Read(&c,1,&read);
              if ( NS_FAILED(rv)) {
				  goto end3;
			  }
				
              if (c == '\r');
              else if (c == '\n') 
                break;
              else {
                  dataHead[j][i]=c;
                  i++;
              }
          }
          if (i == KMAX_HEAD) {
              rv = NS_ERROR_OUT_OF_MEMORY;
			  goto end3;
          }
          dataHead[j][i]='\0';
      }

      LOG(("[Perseus Extension] - \
           [nsPerseusUploadData::readContent] Header : \
           \n%s\n%s",dataHead[0],dataHead[1]));
      rv=mStream->Read(&c,1,&read); //Read
      rv=mStream->Read(&c,1,&read); //Read
      // extract content-type from the header
      
      while(    tmp<KMAX_HEAD
              && dataHead[0][tmp]
              && (dataHead[0][tmp] != ':') )
          tmp++;
      
	  content = (char * ) NS_Alloc(sizeof(char) * KMAX_HEAD);
	  if (!content) {
		rv = NS_ERROR_OUT_OF_MEMORY;
		goto end4;
	  }  
	  
	  tmp +=2;
      while ( tmp<KMAX_HEAD 
              && tmp2<KMAX_HEAD
              && dataHead[0][tmp]
              && (dataHead[0][tmp] != ' ') 
              && (dataHead[0][tmp] != ';')) {
          content[tmp2++]=dataHead[0][tmp++];
      }

      content[tmp2]='\0';
      mContent.Assign(content);
	  rv = NS_OK;
end4:
	  NS_Free(content);
end3:
	  NS_Free(dataHead[1]);
end2:
	  NS_Free(dataHead[0]);
end1:
	  NS_Free(dataHead);
end:
	  return rv;
}

nsresult 
PerseusUploadData::readContentString()
{
	
    LOG(("[Perseus Extension] - [nsPerseusUploadData::readData]"
         "InputStream is a StringInputStream"));
    nsresult rv ;
	NS_NAMED_LITERAL_CSTRING(content,"Content-Type");
	nsCOMPtr<nsIHttpChannel> httpChannel = 
              do_QueryInterface(mChannel, &rv);

    if (NS_FAILED(rv))
	{
	  goto end;
	}
         
    rv=httpChannel->GetRequestHeader(content,mContent);
	if (NS_FAILED(rv)) {
		mContent.Assign("");
	}
	rv = NS_OK;

end :
	  return rv;
} 
