#include "mod_perseus.h"
#include "perseusHTTP.h"

#include "util_filter.h"
#include "apr_buckets.h"
#include "http_protocol.h"
#include "http_request.h"

#define MY_FREE(p)\
        free((p));\
        (p) = NULL;

module AP_MODULE_DECLARE_DATA perseus_module;


typedef struct perseus_ctx_t
{
    unsigned char first_time;
    char * buffer;
    unsigned long bufferLength;
    char * bufferOut;
    unsigned long bufferOutLength;
    request_rec * r;
    PERSEUS_CTYPE_t ctype;
    pcc * code;
    unsigned long read;
    unsigned long toread;
    apr_bucket_brigade * bb; 
    unsigned long bbLength;
    char * boundary;
    unsigned int boundaryLength;
} perseus_ctx;



/*
 * O error
 * 1 more data wait
 * 2 finish
 */
unsigned char
inputMultipartDataProcess(perseus_ctx * ctx, char * data, unsigned long dataLength)
{
	unsigned char ret = 0;
	char * dataDecoded = NULL;
	unsigned long dataDecodedLength = 0;
	unsigned long rest = 0;
	unsigned long read = 0;
	char * cursor = data;
	unsigned long cursorLength = dataLength;


	/* if first time*/
	if (ctx->first_time ==1)
	{
		apr_table_unset(ctx->r->headers_in, "Content-Length");
	//	fprintf(stderr,"first_time : %s !\n",data);
	//	fflush(stderr);
/*		ret = multipartReadPreambule(data,dataLength, &read);
		if ( ret == 0)
		{
			fprintf(stderr,"Preambule error !\n");
			fflush(stderr);
			ctx->bufferOut = addData(&(ctx->bufferOut),&(ctx->bufferOutLength),
                            data,dataLength);
			 goto end;
		}
*/		/* add all data in brigade */
/*		ctx->bufferOut = addData(&(ctx->bufferOut),&(ctx->bufferOutLength),
				data,read);
	//	fprintf(stderr,"=====buffer out : %s !\n======",ctx->bufferOut);
	//	fflush(stderr);
		cursor += read;
		cursorLength -= read;
*/		ret = multipartSearchData(cursor,cursorLength,
				ctx->boundary, ctx->boundaryLength, 
				&read);	
		switch(ret)
		{
			case 1:
				ctx->bufferOut = addData(&(ctx->bufferOut),&(ctx->bufferOutLength),
				cursor,read);
	//			fprintf(stderr,"=====buffer out : %s !\n======",ctx->bufferOut);
	//			fflush(stderr);
				break;
			case 2:
			case 3:
			default:
				ctx->bufferOut = addData(&(ctx->bufferOut),&(ctx->bufferOutLength),
				cursor,cursorLength);
				ret = 1;
				goto end;
				break;
		}
		/* add all data in brigade */
		cursor +=read;
		cursorLength -= read;
		ctx->first_time = 0;
	//	fprintf(stderr,"End first time : %s !\n",cursor);
	//	fflush(stderr);
	}
	/* copy new data in the buffer*/
	ctx->buffer = addData(&(ctx->buffer), &(ctx->bufferLength), cursor, cursorLength);
	do{
		/* decode data */
		ret = perseusHTTPdecode(ctx->code,ctx->buffer,
				ctx->bufferLength,&dataDecoded,&dataDecodedLength,&rest);
	//	fprintf(stderr," decode : %s !\n",dataDecoded);
	//	fflush(stderr);
		if (ret == 0)
		{
			goto end;
		}
		/* add data decoded into brigade */
		ctx->bufferOut = addData(&(ctx->bufferOut),&(ctx->bufferOutLength),
                            dataDecoded,dataDecodedLength);
		MY_FREE(dataDecoded);
		dataDecodedLength = 0;
	//	fprintf(stderr,"buffer out  decode : %s !\n",ctx->bufferOut);
	//	fflush(stderr);
		if (ret == 1)
		{
			/* */
			MY_FREE(ctx->buffer);
			ctx->bufferLength = 0;
			ret =1;
			goto end;
		}

		/* remove data decoded of the buffer */
		ctx->bufferLength -= rest;
		ctx->buffer = memmove(ctx->buffer,ctx->buffer+rest,ctx->bufferLength);
		
		ret = multipartSearchData(ctx->buffer,ctx->bufferLength,
				ctx->boundary, ctx->boundaryLength, 
				&read);	
		switch(ret)
		{
			case 1:
				ctx->bufferOut = addData(&(ctx->bufferOut),&(ctx->bufferOutLength),
				ctx->buffer,read);
				ctx->bufferLength -= read;
				ctx->buffer = memmove(ctx->buffer,ctx->buffer+read,ctx->bufferLength);
				break;
			case 2:
				ret = 1;
				goto end;
				break;
			case 3:
				ctx->bufferOut = addData(&(ctx->bufferOut),&(ctx->bufferOutLength),
				ctx->buffer,read);

				ctx->bufferLength -= read;
				ctx->buffer = memmove(ctx->buffer,ctx->buffer+read,ctx->bufferLength);
				ret = 2;
				goto end;
				break;
			default:
				ret = 0;
				goto end;
				break;
		}

	}while(ctx->bufferLength > 0);	/* loop */
	
	ret = 2;
end: 
	return ret;
}

PERSEUS_CTYPE_t
get_content_type(request_rec * r)
{
    PERSEUS_CTYPE_t ret = PERSEUS_UNSUPPORTED;
    char urlencoded[] = "application/x-www-form-urlencoded";
    unsigned int urlencodedLength = 33;
    char multipart[] =  "multipart/form-data";
    unsigned int multipartLength = 19;
    unsigned int ctypeLength = 0;
    const char * ctype =  apr_table_get(r->headers_in,"Content-Type");
    unsigned int i =0;

    if (ctype == NULL)
    {
        goto end;
    }
    ctypeLength = strlen(ctype);
    /* 
     * Test 
     * application/x-www-form-urlencoded
     */
    if (ctypeLength >= urlencodedLength)
    {
        i=0;
        while ( (i<urlencodedLength) &&  
                (urlencoded[i] == ctype[i]))
                i++;
        if (i==urlencodedLength) {
            ret = PERSEUS_URLENCODED;
            goto end;
        }
    }
    if (ctypeLength >= multipartLength)
    {
        i=0;
        while ( (i<multipartLength) &&  
                (multipart[i] == ctype[i]))
                i++;
        if (i==multipartLength) {
            ret = PERSEUS_MULTIPARTFORMDATA;
            goto end;
        }
    }
end:
    return ret;
}

char *
get_boundary(request_rec * r)
{
	//TODO refaire 
    char * ret = NULL;
    //char boundary[] = "boundary=";
    unsigned int ctypeLength = 0;
    const char * ctype =  apr_table_get(r->headers_in,"Content-Type");
    unsigned int i =0;

    if (ctype == NULL)
    {
        goto end;
    }

    ctypeLength = strlen(ctype);
    /*
     *
     */
    i=0;
    while ( ctype[30+i] != '\r')
        i++;

    
    ret = (char *) apr_palloc(r->pool,sizeof(char)*(i+1+2));
    memcpy(ret+2,ctype+30,i);
    ret[0]='-';
    ret[1]='-';
    ret[i]='\0';

  /*  fprintf(stderr,"boundary= %s\n"
                    "ctype =   %s\n",ret,ctype);
    fflush(stderr);
*/

end:
    return ret;
}

pcc * 
get_pcc(request_rec *r)
{
    server_rec      *srv;
    pcc             *aCode;
    size_t          client_ip_length;
    size_t          perseus_id_length;
    size_t          key_length; 
    const char      *perseus_id;
    unsigned char   *key;
    char            *client_ip;
    char            *message;


    if (r == NULL)
        return NULL;
    if (r->connection == NULL)
        return NULL;
    
    client_ip = r->connection->remote_ip;
    if (client_ip == NULL)
        return NULL;

    client_ip_length = strlen(client_ip);
    
    /*
     * IPv4 max 15 car
     */
    if (client_ip_length > 16)
        return NULL;

    srv = r->server;

    if (srv == NULL)
        return NULL;

    perseus_id = apr_table_get(r->headers_in, "Perseus_id");

    if (perseus_id == NULL) 
        return NULL;
    
    perseus_id_length = strlen(perseus_id);
    
    /*
     * MAX_INT = 2147483647 (10 char)
     */
    if (perseus_id_length > 10)
        return NULL;

    key_length = perseus_id_length + client_ip_length; 
    key = (unsigned char *) 
        apr_palloc(r->pool, sizeof(unsigned char) * (key_length+1));        
    
    if (key == NULL)
        return NULL;

    memcpy(key, client_ip, client_ip_length);
    memcpy(key+client_ip_length, perseus_id, perseus_id_length);
    key[key_length] = '\0';

    ap_log_error(APLOG_MARK,APLOG_NOTICE,0,srv,
            "Perseus hash key : %s ",key);

    message = perseus_shmht_retrieve(srv, key, key_length);
    if (message == NULL)
    {
        ap_log_error(APLOG_MARK,APLOG_NOTICE,0,srv,
                "ERROR Perseus hash key unkonwn : %s ",key);
        return NULL;
    }
    
    //aCode=pcc_save(r->pool,message);
    aCode = parsePCC(message,strlen(message));
    if (aCode == NULL) {
        ap_log_error(APLOG_MARK,APLOG_NOTICE,0,srv,
                    "ERROR Perseus pcc");
    }
        ap_log_error(APLOG_MARK,APLOG_NOTICE,0,srv,
                    "End get pcc");
    return aCode;
}

apr_status_t 
set_pcc(request_rec *r, char *param)
{
    server_rec      *srv;
    apr_time_t      expiry;
    size_t          client_ip_length;
    size_t          perseus_id_length;
    size_t          key_length;
    apr_status_t    ret;
    const char      *perseus_id;
    unsigned char   *key; 
    char            *client_ip;
   
    if (r == NULL)
        return APR_EGENERAL;
    if (r->connection == NULL)
        return APR_EGENERAL;
    
    client_ip = r->connection->remote_ip;
    
    if (client_ip == NULL)
        return APR_EGENERAL;

    client_ip_length = strlen(client_ip);

    /*
     * IPv4 max 15 car
     */
    if (client_ip_length > 16)
        return APR_EGENERAL;

    srv = r->server;

    if (srv == NULL)
        return APR_EGENERAL;

    perseus_id = apr_table_get(r->headers_in, "Perseus_id");
    if (!perseus_id)
        return APR_EGENERAL;
    
 
    perseus_id_length = strlen(perseus_id);

    /*
     * MAX_INT = 2147483647 (10 char)
     */
    if (perseus_id_length > 10)
        return APR_EGENERAL;


    key_length = perseus_id_length + client_ip_length;
 
    key = (unsigned char *) 
        apr_palloc(r->pool, sizeof(unsigned char) * (key_length+1));        

    memcpy(key, client_ip, client_ip_length);
    memcpy(key+client_ip_length, perseus_id, perseus_id_length);
    key[key_length] = '\0';

    ap_log_error(APLOG_MARK,APLOG_NOTICE,0,srv,
            "Perseus hash key : %s ",key);
    
    expiry = apr_time_now() + PERSEUS_TIMEOUT;
    ret = perseus_shmht_store(srv, key,key_length, expiry, param);
   
    return ret;
}


static void *
create_perseus_config(apr_pool_t *p, server_rec *s)
{
    perseus_config *c;
    
    c = (perseus_config *) apr_pcalloc(p, sizeof(perseus_config));
    
    if (!c) {
        ap_log_error(APLOG_MARK, APLOG_CRIT, 0, s,
                "Failed to allocate memory for perseus_config");
        exit(1);
    }

    c->shm_filename = (char *) apr_palloc(p,sizeof(char)*17);
    
    if (!c->shm_filename) {
        ap_log_error(APLOG_MARK, APLOG_CRIT, 0, s,
                "Failed to allocate memory or shm_filename");
         exit(1);
    }
    
    apr_cpystrn(c->shm_filename, SHM_FILE, SHM_FILE_L);

    c->shm_size = 1024 * 512; /* 512 KB */

    c->cMutexFilename = (char *) apr_palloc(p,sizeof(char)*18);
    if (!c->cMutexFilename) {
        ap_log_error(APLOG_MARK, APLOG_CRIT, 0, s,
                "Failed to allocate memory for cMutexFilename");
         exit(1);
    }
    apr_cpystrn(c->cMutexFilename,LOCK_FILE,LOCK_FILE_L);
    
    return c;
}
/*
 * static void *
 * merge_perseus_config(apr_pool_t *p, void *basev, void *overridesv)
 * {
 *   return NULL;
 * }  
 */

int
util_read (request_rec * r, char **rbuf)
{
  
  unsigned long length;
  int           rc; 
  int           rsize;
  size_t        len_read;
  size_t        rpos;
  char          argsbuffer[HUGE_STRING_LEN];
  
  rc = OK;

  if ((rc = ap_setup_client_block (r, REQUEST_CHUNKED_ERROR)) != OK)
    return rc;

  if (ap_should_client_block (r))
  {
      rpos = 0;
      length = r->remaining;

      *rbuf = apr_pcalloc (r->pool, length + 1);

      while ((len_read =
              ap_get_client_block (r, argsbuffer, sizeof (argsbuffer))) > 0)
      {
          if ((rpos + len_read) > length)
              rsize = length - rpos;
          else
              rsize = len_read;
          memcpy ((char *) *rbuf + rpos, argsbuffer, rsize);
          rpos += rsize;
      }
  }

  return rc;
}

/*
 * Decode data for a GET method
 */
static int 
perseus_hook_method (request_rec *r)
{

	const char * perseus_type = apr_table_get(r->headers_in, "Perseus");

	if (!perseus_type)
		return DECLINED;

	if (!apr_strnatcmp(perseus_type,"pcc")) {

		if (r->method_number == M_GET) {
			char * args = r->args;
			pcc * sess; 
			unsigned int args_len = strlen(args);
			char * dataDecode;
			unsigned long long decodedLength;
			unsigned int i;
			
			apr_table_unset(r->headers_in, "Perseus");

			sess=get_pcc(r);
			if (sess == NULL)
			{
				fprintf(stderr,"Error to get PCC\n");
				fflush(stderr);
				return DECLINED;
			}

			dataDecode = parseGet(args,args_len,sess,PERSEUS_DECODE);
			if (dataDecode == NULL)
			{
				fprintf(stderr,"Error to decode!");
				fflush(stderr);
				return DECLINED;
			}
			decodedLength = strlen(dataDecode);
			
			fprintf(stderr,"\n Data decoded [%llu] : %s \n", decodedLength, dataDecode );
			fflush(stderr);
			for (i=0; i<decodedLength; i++)
			{
				r->args[i] = dataDecode[i];
			}
			r->args[i]='\0';
			apr_table_set(r->headers_out, "Perseus","PCC");
			/* Add ID */
		}
	}

	return DECLINED;
}

/*Handler allow to communicate with the client*/
static int perseus_handler (request_rec *r)
{
    server_rec  *s = r->server;
    pcc * sess; 
    apr_status_t ret;
    char * message;
    const char * perseus_type = apr_table_get(r->headers_in, "Perseus");
    get_content_type(r);

    if (!perseus_type)
        return DECLINED;
    
    fprintf(stderr,"Perseus handler, Perseus header found : %s\n",perseus_type);
    /*
     * Client ask if we have Perseus Module
     */
    if (!apr_strnatcmp(perseus_type,"check")) {
        /*
         * We can verify the method => GET
         */
        fprintf(stderr,"Perseus handler, header ack found \n");
        fflush(stderr);
        apr_table_set(r->headers_out,"Perseus","check-ack");
        return OK;
    }

    /*
     * Client ask if we know the parameters
     */
    if (!apr_strnatcmp(perseus_type,"known")) {
        /*
         * We can verify the method => GET
         */
        fprintf(stderr,"Perseus handler, header known found \n");
        fflush(stderr);
        sess = get_pcc(r);
        if (sess == NULL)
        {
            apr_table_set(r->headers_out,"perseus","known-no");
        } else {
            apr_table_set(r->headers_out,"perseus","known-yes");
        }
        /*
         * free(sess);
         */
        return OK;
    }

    /*
     * We receive the Perseus parameters
     */
    if (!apr_strnatcmp(perseus_type,"init")) {
        /*
         * We can verify the method => POST
         */
        fprintf(stderr,"Perseus handler, header init found \n");
        fflush(stderr);

        message = NULL;
        if ( util_read (r, &message) != OK )
        {
            ap_log_error(APLOG_MARK,APLOG_NOTICE,0,s,
                    "Error read message"); 
            apr_table_set(r->headers_out, "Perseus","init-nack");
            return OK;
        }

        ret = set_pcc(r,message);
        if ( ret != APR_SUCCESS ) {
            apr_table_set(r->headers_out, "Perseus","init-nack");
            return OK;
        }
        /*
         * We must add the parameters in the table
         */
        apr_table_set(r->headers_out, "Perseus","init-ack");
        return OK;
    }
    
    return DECLINED;
}

static apr_status_t
init_context(request_rec * r, perseus_ctx ** ctx)
{
    apr_status_t ret;
    perseus_ctx *ctx_tmp = *ctx;
    ctx_tmp = apr_pcalloc(r->pool, sizeof(perseus_ctx));
    if (ctx_tmp == NULL)
    {
        ret = APR_EGENERAL;
        goto end;
    }
    
    ctx_tmp->r = r;
    ctx_tmp->first_time = 1;
    ctx_tmp->buffer = NULL;
    ctx_tmp->bufferLength = 0;
    ctx_tmp->ctype = get_content_type(r);

    if (ctx_tmp->ctype == PERSEUS_UNSUPPORTED)
    {
        ret = APR_EGENERAL;
        goto end;
    }    

    if (ctx_tmp->ctype == PERSEUS_MULTIPARTFORMDATA)
    {

        ctx_tmp->boundary = get_boundary(r);
        if (ctx_tmp->boundary == NULL)
        {
            ret = APR_EGENERAL;
            goto end;
        }
        ctx_tmp->boundaryLength = strlen(ctx_tmp->boundary);
    }
    else
    {
        ctx_tmp->boundary = NULL;
        ctx_tmp->boundaryLength = 0;
    }

    ctx_tmp->code = get_pcc(r);
    if (ctx_tmp->code == NULL) 
    {
        ret = APR_EGENERAL;
        goto end;
    }
    const char * length = apr_table_get(r->headers_in, "Content-Length");
    ctx_tmp->toread = atol(length);
    ctx_tmp->bb = apr_brigade_create(r->pool, r->connection->bucket_alloc);
    if (ctx_tmp->bb == NULL)
    {
        ret = APR_EGENERAL;
        goto end;
    }
    ctx_tmp->bbLength = 0;
    ctx_tmp->read = 0;
    *ctx=ctx_tmp;
    ret = APR_SUCCESS;
    ctx_tmp->bufferOut = NULL;
    ctx_tmp->bufferOutLength = 0;
end:
    return ret;
}

static apr_status_t
init_contextiout(request_rec * r, perseus_ctx ** ctx)
{
    apr_status_t ret;
    perseus_ctx *ctx_tmp = *ctx;
    ctx_tmp = apr_pcalloc(r->pool, sizeof(perseus_ctx));
    if (ctx_tmp == NULL)
    {
        ret = APR_EGENERAL;
        goto end;
    }
    
    ctx_tmp->r = r;
    ctx_tmp->first_time = 1;
    ctx_tmp->buffer = NULL;
    ctx_tmp->bufferLength = 0;

    ctx_tmp->code = get_pcc(r);
    if (ctx_tmp->code == NULL) 
    {
        ret = APR_EGENERAL;
        goto end;
    }
    ctx_tmp->bb = apr_brigade_create(r->pool, r->connection->bucket_alloc);
    if (ctx_tmp->bb == NULL)
    {
        ret = APR_EGENERAL;
        goto end;
    }
    ctx_tmp->bbLength = 0;
    ctx_tmp->read = 0;
    *ctx=ctx_tmp;
    ctx_tmp->bufferOut = NULL;
    ctx_tmp->bufferOutLength = 0;
    ret = APR_SUCCESS;
end:
    return ret;
}

/*
 * Input  Filter allows to decode a stream
 * BETA VERSION
 */ 
static apr_status_t perseus_in_filter(ap_filter_t *f,
                                      apr_bucket_brigade *bb,
                                      ap_input_mode_t mode,
                                      apr_read_type_e block,
                                      apr_off_t readbytes)
{

    apr_status_t ret;
    unsigned char ret2;
    request_rec *r = f->r;
    perseus_ctx * ctx = f->ctx;
    const char *perseus = NULL;
    const char * data = NULL;
    apr_size_t dataLength = 0;
    apr_bucket *bkt = NULL; 
    apr_bucket_brigade *bbinp = NULL;


    /*
     * just get out of the way of things we don't want.
     */
    if (r->main) {
        ap_remove_input_filter(f);
        return ap_get_brigade(f->next, bb, mode, block, readbytes);
    }
    if (mode != AP_MODE_READBYTES) {
        ap_remove_input_filter(f);
        return ap_get_brigade(f->next, bb, mode, block, readbytes);
    }

    if (f->c->aborted) {
        ap_log_rerror(APLOG_MARK, APLOG_NOTICE, 0, r,
                "Perseus aborted");
        bkt = apr_bucket_eos_create(f->c->bucket_alloc);
        APR_BRIGADE_INSERT_TAIL(bb, bkt);
        return APR_ECONNABORTED;
    }

    if (r->method_number == M_GET) {
        ap_remove_input_filter(f);
        return ap_get_brigade(f->next, bb, mode, block, readbytes);
    }   
    
    /* Check perseus headers */
    perseus = apr_table_get(r->headers_in, "Perseus");

    if (!perseus) {
        ap_remove_input_filter(f);
        return ap_get_brigade(f->next, bb, mode, block, readbytes);
    }
    if (apr_strnatcmp(perseus,"pcc")) { 
        ap_remove_input_filter(f);
        return ap_get_brigade(f->next, bb, mode, block, readbytes);
    } 

    if (ctx == NULL)
    {
        ret = init_context(r, &ctx);
        if (ret != APR_SUCCESS)
        {
            return ret;
        }

        PERSEUS_CTYPE_t ctype = get_content_type(r);
        if (ctype == PERSEUS_UNSUPPORTED) {
            ap_remove_input_filter(f);
            return ap_get_brigade(f->next, bb, mode, block, readbytes);
        }    
        f->ctx=ctx;
	apr_table_set(r->headers_out, "Perseus","PCC");
	fprintf(stderr,"Perseus Input Filter begin \n");
	fflush(stderr);
    }


    bbinp = apr_brigade_create(f->r->pool, f->c->bucket_alloc);
    while (APR_BRIGADE_EMPTY(ctx->bb))
    { 
	   ret = ap_get_brigade(f->next, bbinp, mode, block, readbytes);
	   if (ret != APR_SUCCESS) 
	   {
		   ap_log_rerror(APLOG_MARK, APLOG_NOTICE, 0, r,
				   "Error in ap_get_brigade!");
		   return ret;
	   }

	   for (bkt = APR_BRIGADE_FIRST(bbinp);
			   bkt != APR_BRIGADE_SENTINEL(bbinp);
			   bkt = APR_BUCKET_NEXT(bkt))
	   {
		   if (APR_BUCKET_IS_EOS(bkt)) {
			   /*
			    * ap_log_rerror(APLOG_MARK, APLOG_NOTICE, 0, r,
					   "bkt is EOS decode");
			    */
			   if (ctx->bufferLength != 0)
			   {
				   ap_log_rerror(APLOG_MARK, APLOG_NOTICE, 0, r,
						   "bkt is EOS but buffer not empty ");
				   /*Try */
				   //ret2 =inputMultipartDataProcess(ctx,"",0);
			   }
			   if (ctx->bufferOutLength != 0)
			   {
	/*			   fprintf(stderr,"\n=============\nEnd filter[%lu] : %s\n=============\n",ctx->bufferOutLength,ctx->bufferOut);
				   fflush(stderr);
	*/			   /*    ret = ap_fwrite(r->input_filters,
					 bb,
					 ctx->bufferOut,
					 ctx->bufferOutLength);
					 */
				   bkt = apr_bucket_heap_create((char *)ctx->bufferOut,ctx->bufferOutLength,NULL,f->c->bucket_alloc);
				   MY_FREE(ctx->bufferOut);
				   ctx->bufferOutLength = 0;
				   //	    APR_BRIGADE_CONCAT(bb, ctx->bb);
				   APR_BRIGADE_INSERT_TAIL(bb,bkt);
			   }
			   else{
				   //APR_BUCKET_REMOVE(bkt);
				   bkt = apr_bucket_eos_create(f->c->bucket_alloc);
				   // APR_BRIGADE_CONCAT(bb, ctx->bb);
				   APR_BRIGADE_INSERT_TAIL(bb, bkt);
			   }
			   return APR_SUCCESS;
		   }
		   if (APR_BUCKET_IS_FLUSH(bkt)) {
			   ap_log_rerror(APLOG_MARK, APLOG_NOTICE, 0, r,
					   "bkt is a flush");
			   continue;
		   }

		   /* read */

		   ret = apr_bucket_read(bkt, &data, &dataLength, APR_BLOCK_READ);
		   if (ret == APR_SUCCESS) 
		   {
			   	//	    fprintf(stderr,"data read : %s\n",data);
			   	//	    fflush(stderr);
			   ctx->read = ctx->read + dataLength;
			   /* O error
			    * 1 more data wait
			    * 2 finish
			    */
			   switch(ctx->ctype)
			   {
				   case PERSEUS_MULTIPARTFORMDATA:
					   {
						   ret2 =inputMultipartDataProcess(ctx,(char *)data,dataLength);
						   if (ret2 == 0)
						   {
							   ap_log_rerror(APLOG_MARK, APLOG_NOTICE, 0, r,
									   "Error in inputMultipartDataProcess !");
							   return APR_EGENERAL;
						   }
						   dataLength = 0;
						   break;
					   }
					case PERSEUS_URLENCODED:
					   {
						   apr_table_unset(ctx->r->headers_in, "Content-Length");
						   ctx->bufferOut = parseGet((char *)data,dataLength,ctx->code,PERSEUS_DECODE);
						   ctx->bufferOutLength = strlen(ctx->bufferOut);

						   break;
					   }
					default:
 
					   return APR_EGENERAL;
			   }
					   
		   }

	   }

	   apr_brigade_cleanup(bbinp);
	   apr_brigade_destroy(bbinp);
	   if (ctx->bufferOutLength != 0)
	   {
	/*	   fprintf(stderr,"\n=============\nEnd filter : %s\n=============\n",ctx->bufferOut);
		   fflush(stderr);
	*/	   /* ret = ap_fwrite(r->input_filters,
		      bb,
		      ctx->bufferOut,
		      ctx->bufferOutLength);
		      */
		   bkt = apr_bucket_heap_create((char *)ctx->bufferOut,ctx->bufferOutLength,NULL,f->c->bucket_alloc);
		   APR_BRIGADE_INSERT_TAIL(ctx->bb,bkt);
		   //APR_BRIGADE_CONCAT(bb, ctx->bb);
		   MY_FREE(ctx->bufferOut);
		   ctx->bufferOutLength = 0;

		   /*	    if (ret != APR_SUCCESS)
			    {
			    fprintf(stderr,"\nError write\n");
			    return APR_EGENERAL;
			    }
			    */
	   }   
	   else{

	/*	   fprintf(stderr,"\n======VIDE VIDE========\n");
		   fflush(stderr);
	*/	   /*	    ret = ap_get_brigade(f->next, bbinp, mode, block, readbytes);
			    if (ret != APR_SUCCESS) 
			    {
			    ap_log_rerror(APLOG_MARK, APLOG_NOTICE, 0, r,
			    "Error in ap_get_brigade!");
			    return ret;
			    }
			    fprintf(stderr,"\n======MORE DISPO========\n");
			    fflush(stderr);
			    */

	   }
   }
   /*if (ctx->read == ctx->toread)
    {
	    ap_log_rerror(APLOG_MARK, APLOG_NOTICE, 0, r,
			    "It's the end of stream!\n");

	    if (ctx->bufferLength !=0)
	    {
		    ap_log_rerror(APLOG_MARK, APLOG_NOTICE, 0, r,
				    "No more data and buffer not empry : %lu !",ctx->bufferLength);
	    }

	    APR_BUCKET_REMOVE(bkt);
	    bkt = apr_bucket_eos_create(f->c->bucket_alloc);
	    //APR_BRIGADE_CONCAT(bb, ctx->bb);
	    APR_BRIGADE_INSERT_TAIL(bb, bkt);

    }
    */
    if (!APR_BRIGADE_EMPTY(ctx->bb))
    {
	    apr_bucket_brigade *newbb;

	    apr_brigade_partition(ctx->bb,readbytes,&bkt);

	    newbb = apr_brigade_split(ctx->bb,bkt);

	    APR_BRIGADE_CONCAT(bb,ctx->bb);
	    APR_BRIGADE_CONCAT(ctx->bb,newbb);

    }
    return APR_SUCCESS;
}

/*
 * ALPHA VERSION 
 */
static apr_status_t perseus_out_filter(ap_filter_t *f,
        apr_bucket_brigade *bb)

{
	apr_status_t ret = APR_SUCCESS;
	request_rec *r = f->r;
	apr_bucket *bkt;
	apr_read_type_e rblock = APR_NONBLOCK_READ;
	perseus_ctx * ctx = f->ctx;
	const char * data = NULL;
	apr_size_t dataLength = 0;

	const char * perseus; 
	perseus = apr_table_get(r->headers_out, "Perseus");
	if (!perseus) {
		ap_remove_input_filter(f);
		return ap_pass_brigade(f->next, bb);
	}

	fprintf(stderr,"OUTPUT FILTER Header Perseus found : [%s] \n",perseus);
	fflush(stderr);
	if (apr_strnatcmp(perseus,"PCC")) { 
		ap_remove_input_filter(f);
		return ap_pass_brigade(f->next, bb);
	}    

	if (!ctx) {
		if ( r->main) {
			ap_remove_output_filter(f);
			return ap_pass_brigade(f->next, bb);
		}

		/* Do nothing if asked to filter nothing. */
		if (APR_BRIGADE_EMPTY(bb)) {
			ap_remove_output_filter(f);
			return ap_pass_brigade(f->next, bb);
		}
		if (!ap_is_initial_req(r)) {
			ap_remove_output_filter(f);
			return ap_pass_brigade(f->next, bb);
		}

		ret = init_contextiout(r, &ctx);
		if (ret != APR_SUCCESS)
		{
			return ret;
		}

		f->ctx=ctx;
	}  

	for (bkt = APR_BRIGADE_FIRST(bb);
			bkt != APR_BRIGADE_SENTINEL(bb);
			bkt = APR_BUCKET_NEXT(bkt))
	{
        if (APR_BUCKET_IS_EOS(bkt))
        {
            if (ctx->bufferOutLength != 0)
            {
                /*
                 * fprintf(stderr,"\n=============\nEnd filter[%lu] : %s\n=============\n",ctx->bufferOutLength,ctx->bufferOut);
                 *				fflush(stderr);
                 */
				
				bkt = apr_bucket_heap_create((char *)ctx->bufferOut,ctx->bufferOutLength,NULL,f->c->bucket_alloc);
				APR_BRIGADE_INSERT_TAIL(ctx->bb,bkt);
				MY_FREE(ctx->bufferOut);
				ctx->bufferOutLength = 0;
			}

			bkt = apr_bucket_eos_create(f->c->bucket_alloc);
			APR_BRIGADE_INSERT_TAIL(ctx->bb, bkt);
			return ap_pass_brigade(f->next,ctx->bb);
		}
		if (APR_BUCKET_IS_FLUSH(bkt)) {
			ap_log_rerror(APLOG_MARK, APLOG_NOTICE, 0, r,
					"bkt is a flush");
			continue;
		}

		ret = apr_bucket_read(bkt, &data, &dataLength, rblock);
		if (ret == APR_SUCCESS) 
		{
				//	    fprintf(stderr,"data read : %s\n",data);
				//	    fflush(stderr);
			ctx->read = ctx->read + dataLength;
			/* Add data to buffer */
			perseusHTTPencode(ctx->code,(char *)data,dataLength,
					&(ctx->bufferOut), &(ctx->bufferOutLength));
		}
	}

	fprintf(stderr,"Exit Ouput Filter\n");
	fflush(stderr);
	return ap_pass_brigade(f->next,ctx->bb);
}

apr_status_t
cleanup_resources(void * s)
{

    server_rec *base_server = (server_rec *) s;
    ap_log_error(APLOG_MARK,APLOG_NOTICE,0,0,"Perseus cleanup resources");

    perseus_shmht_kill(base_server); 
    return APR_SUCCESS;
}


int perseus_init_Module(apr_pool_t *p, apr_pool_t *plog,
        apr_pool_t *ptemp,
        server_rec *base_server)
{
    void *data;
    const char *userdata_key = "perseus_init";
    apr_status_t ret;

    perseus_config *c = (perseus_config *)
        ap_get_module_config(base_server->module_config, &perseus_module);

    plog = plog;
    ptemp = ptemp;
    ap_log_error(APLOG_MARK,APLOG_NOTICE,0,base_server,"Perseus init module");
    if (!c) 
    {
        ap_log_error(APLOG_MARK, APLOG_CRIT, 0, base_server,
                "Failed to get perseus_config");
        return HTTP_INTERNAL_SERVER_ERROR;
    }

    /* initialize_module() will be called twice, and if it's a DSO
     * then all static data from the first call will be lost. Only
     * set up our static data on the second call. */
    apr_pool_userdata_get(&data, userdata_key, base_server->process->pool);
    if (!data) {
        apr_pool_userdata_set((const void *)1, userdata_key,
                apr_pool_cleanup_null, base_server->process->pool);
        return OK;
    }

    ret = perseus_shmht_init(base_server,p);
    if ( ret != APR_SUCCESS)
        return HTTP_INTERNAL_SERVER_ERROR;

    /* Register a cleanup function */
    apr_pool_cleanup_register(p, base_server,
            cleanup_resources,
            apr_pool_cleanup_null);

    return 0;
}


static void register_hooks(apr_pool_t *p)
{
    //p = p;
    ap_hook_post_config(perseus_init_Module, NULL,NULL, APR_HOOK_MIDDLE);

    ap_hook_translate_name(perseus_hook_method, NULL, NULL, APR_HOOK_REALLY_FIRST);
    ap_hook_handler(perseus_handler, NULL, NULL,APR_HOOK_REALLY_FIRST);
    ap_register_input_filter(PerseusFilterName, perseus_in_filter, NULL,AP_FTYPE_RESOURCE);
		    //AP_FTYPE_TRANSCODE);
    //        AP_FTYPE_CONTENT_SET);

    //AP_FTYPE_CONNECTION+4);
    ap_register_output_filter(PerseusFilterName, perseus_out_filter, NULL,
            AP_FTYPE_CONTENT_SET-1 );
    return;
}


module AP_MODULE_DECLARE_DATA perseus_module = {
    STANDARD20_MODULE_STUFF,
    NULL,                           /* dir config creater */
    NULL,                           /* dir merger --- default is to override */
    create_perseus_config,          /* server config */
    NULL,           /* merge server config */
    NULL,                           /* command table */
    register_hooks                  /* register hooks */
};
