/* 
	Module Perseus permet la gestion des trames codes
	avec les Codes Convolutifs Poinonnes
*/

#ifndef __MOD_PERSEUS_H__
#define __MOD_PERSEUS_H__

#include "httpd.h"
#include "http_config.h"
#include "http_log.h"

#include "apr_shm.h"
#include "apr_rmm.h"
#include "apr_strings.h"


#include "apr_global_mutex.h"
#ifdef AP_NEED_SET_MUTEX_PERMS
#include "unixd.h"
#endif

#include "apr_general.h"
#include "util_table.h"

#define PERSEUS_TIMEOUT  300000000
 
#include "perseus.h"
#include "perseusHTTP.h"

#define mySrvConfig(srv) (perseus_config *)ap_get_module_config(srv->module_config,  &perseus_module)

static const char PerseusFilterName[] = "PERSEUS";
module AP_MODULE_DECLARE_DATA perseus_module;

#ifdef _WIN32
#define SHM_FILE "c:/perseus.shm"
#define SHM_FILE_L 14 
#define LOCK_FILE "c:/perseus.lock"
#define LOCK_FILE_L 15 
#else
#define SHM_FILE  "/tmp/perseus.shm"
#define SHM_FILE_L 17
#define LOCK_FILE "/tmp/perseus.lock"
#define LOCK_FILE_L 18 
#endif

/* Struct */
/*
	Structure de configuration du module
*/
typedef struct perseus_config_t
{
    char * shm_filename;
    apr_shm_t * perseus_shm; /* Shared mem */
    apr_size_t shm_size;   
	apr_rmm_t * perseus_rmm;
    table_t * perseus_hashtable; /*Hash Table */
    apr_global_mutex_t *pMutex; /* Mutex */
    char * cMutexFilename;  /* Mutex Name*/
    int session_cache_timeout;
}perseus_config;

/* API */
/*
	Session Cache Support 
	Gestion de la table de hash dans
	la mmoire partage.
*/
apr_status_t perseus_shmht_init(server_rec *, apr_pool_t *);
void         perseus_shmht_kill(server_rec *);
apr_status_t perseus_shmht_store(server_rec *, unsigned char *, int, apr_time_t, char *);
char *        perseus_shmht_retrieve(server_rec *, unsigned char *, int);
void         perseus_shmht_remove(server_rec *, unsigned char *, int);
void         perseus_shmht_expire(server_rec *);

/*Mutex Support */
int perseus_mutex_on(server_rec*);
int perseus_mutex_off(server_rec*);


#endif /* __MOD_PERSEUS_H__*/
