#include "nsPerseusDownloadData.h"

#include "nsIHttpChannel.h" 
#include "nsStringAPI.h"
#include "nsIURI.h" 

#include "nsIStorageStream.h"
#include "nsStorageStream.h" 
#include "nsIBinaryOutputStream.h"
#include "nsIBinaryInputStream.h"

#include "prmem.h"


NS_IMPL_ISUPPORTS2(PerseusDownloadData, nsIStreamListener,
                    nsIRequestObserver)

/*****************************************************************************\
 *
 * Class nsPerseusDownloadData Implementation
 *
 *
 * ***************************************************************************/

PerseusDownloadData::PerseusDownloadData(nsPerseusCode *aCode) 
{
 LOG(("[Perseus Extension] - \
       [nsPerseusDownloadData::~nsPerseusDownloadData]"));
  mCode=aCode;
}

PerseusDownloadData::~PerseusDownloadData()
{
  LOG(("[Perseus Extension] - \
       [nsPerseusDownloadData::~nsPerseusDownloadData]"));
  //NS_Free(mData);
}


/*****************************************************************************\
 *
 * nsIRequestObserver Implementation
 *
 *
 * ***************************************************************************/

/*Avant le dbut du tlchargement des donnes la fonction
	OnStartRequest est appele, permet l'initialisation des donnes */
NS_IMETHODIMP
PerseusDownloadData::OnStartRequest(nsIRequest* aRequest,
                              nsISupports* aContext)
{
   LOG(("[Perseus Extension] - \
       [nsPerseusDownloadData::~OnStartRequest]"));

 /*Initialise data */ 
  mRead=0;
  mData=0;
  mError=0;
  return NS_OK;
}

/* 
		A la fin du tlchargement La fonction OnStopRequest est appele
		permet le dchiffrement des donnes et de redonner la main 
		 firefox avec les donnes dcodes
*/
NS_IMETHODIMP
PerseusDownloadData::OnStopRequest(nsIRequest* aRequest,
                             nsISupports* aContext,
                             nsresult aStatus)
{
  LOG(("[Perseus Extension] - \
       [nsPerseusDownloadData::~OnStopRequest]"));

  nsresult rv;
 if (mError) {
    mListener->OnStopRequest(aRequest,aContext,aStatus);
    return  NS_ERROR_FAILURE; 
 }
 nsCOMPtr<nsIChannel> channel = do_QueryInterface(aRequest,&rv);
  NS_ENSURE_SUCCESS(rv, rv);

  if (!channel)
    return NS_ERROR_FAILURE;
  nsCAutoString host;
  
  nsCOMPtr<nsIURI> Uri ;
  rv=channel->GetURI(getter_AddRefs(Uri));
  NS_ENSURE_SUCCESS(rv, rv);

  rv=Uri->GetHost(host);
  NS_ENSURE_SUCCESS(rv, rv);

  nsCAutoString datadecoded;
  nsCAutoString encoded;
 
  if (mRead !=0)
  {
      mData[mRead]='\0';
  }
  encoded.Assign(mData);
  LOG(("[Perseus Extension] - \
       [nsPerseusDownloadData::~OnStopRequest] : Data encoded : %s",mData)); 
  /* Data are decoded */
  mCode->decode(host,encoded,encoded.Length(),datadecoded);
  LOG(("[Perseus Extension] - "
       "[nsPerseusDownloadData::~OnStopRequest] : Data decoded "
       "received :  %s",datadecoded.get()));


  nsCOMPtr<nsIStorageStream> storageStream =  
    do_CreateInstance(NS_STORAGESTREAM_CONTRACTID, &rv);
  NS_ENSURE_SUCCESS(rv, rv);

  rv=storageStream->Init(8192,mRead,nsnull);
  NS_ENSURE_SUCCESS(rv, rv);

  nsCOMPtr<nsIBinaryOutputStream> binaryOutStream = 
    do_CreateInstance("@mozilla.org/binaryoutputstream;1", &rv);
  NS_ENSURE_SUCCESS(rv, rv);

  nsCOMPtr<nsIOutputStream> out;
  rv=storageStream->GetOutputStream(0,getter_AddRefs(out));
  NS_ENSURE_SUCCESS(rv, rv);

  rv=binaryOutStream->SetOutputStream(out);
  NS_ENSURE_SUCCESS(rv, rv);

  rv= binaryOutStream->WriteBytes(datadecoded.get(), datadecoded.Length());
  NS_ENSURE_SUCCESS(rv, rv);


  nsCOMPtr<nsIInputStream> inputStream ;
  rv=storageStream->NewInputStream(0,getter_AddRefs(inputStream));
  NS_ENSURE_SUCCESS(rv, rv);
  
  mListener->OnStartRequest(aRequest,aContext);

  mListener->OnDataAvailable(aRequest,aContext,inputStream,0,
                             datadecoded.Length());

  mListener->OnStopRequest(aRequest,aContext,aStatus);
 
  return NS_OK;
}


/*****************************************************************************\
 *
 * nsIStreamListener Implementation
 *
 *
 * ***************************************************************************/

/*Fonction de tlchargement des donnes codes, peut-tre appele plusieurs fois
	en fonction la taille des donnes, les donnes tlcharges sont stockes dans
		le buffer mData */
NS_IMETHODIMP
PerseusDownloadData::OnDataAvailable(nsIRequest* aRequest,
                                     nsISupports* aContext,
                                     nsIInputStream *aIStream,
                                     PRUint32 aSourceOffset,
                                     PRUint32 aLength)
{
  LOG(("[Perseus Extension] - \
       [nsPerseusDownloadData::~OnDataAvailable]"));
  
  nsresult rv;
  char * data;

  nsCOMPtr<nsIBinaryInputStream> binaryInStream = 
    do_CreateInstance("@mozilla.org/binaryinputstream;1", &rv);
  NS_ENSURE_SUCCESS(rv, rv);
  
  rv=binaryInStream->SetInputStream(aIStream);
  NS_ENSURE_SUCCESS(rv, rv);

  mData= (char *) NS_Realloc((void *) mData,sizeof(char) * (aLength+mRead) );
  if (!mData) {
      mError=1;
      mListener->OnDataAvailable(aRequest,aContext,aIStream,aSourceOffset,
                                 aLength);
      rv = NS_ERROR_FAILURE; 
	  goto end;
  }

  data = (char *) NS_Alloc(sizeof(char)*aLength); 
  if (!data) {
      mListener->OnDataAvailable(aRequest,aContext,aIStream,aSourceOffset,
                                 aLength);
      mError = 1;
      rv = NS_ERROR_FAILURE; 
	  goto end;
  }

  rv=binaryInStream->ReadBytes(aLength,&data);
  if (NS_FAILED(rv))
  {
	  mError = 1;
	  goto end1;
  }
  LOG(("Data downloaded : %s",data)); 
  memcpy(mData+mRead,data,aLength);
  mRead+=aLength;

  LOG(("[Perseus Extension] - "
       "[nsPerseusDownloadData::~OnDataAvailable] mRead : %lu",mRead));

  rv = NS_OK;

end1:
	NS_Free(data);
end:
  return rv;
} 

